# pylint: disable=W0622
"""cubicweb-seo application packaging information"""

modname = "cubicweb_seo"
distname = "cubicweb-seo"

numversion = (1, 0, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "search engine optimisation with robotstxt and sitemaps"
web = f"https://forge.extranet.logilab.fr/cubicweb/cubes/{distname}"

__depends__ = {
    "cubicweb": ">= 4.0.0, < 5.0.0",
    "cubicweb-web": ">= 1.0.0, < 2.0.0",
}
__recommends__ = {}

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: JavaScript",
]



# copyright 2017-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for cubicweb-seo views."""

from io import BytesIO
from datetime import date, datetime
import gzip
import os
import tempfile
import zipfile

from cubicweb_web.devtools.testlib import WebCWTC
from lxml import etree

from cubicweb_seo.views import SitemapRule

SITEMAP_NS = "{http://www.sitemaps.org/schemas/sitemap/0.9}"


def _sitemap_tag(tag_name):
    """Return the complete tag_name in the Sitemap XML namespace from the given short one."""
    return f"{SITEMAP_NS}{tag_name}"


#
# Test sitemap rules for the test schema.py in test/data
#

_TITLEDTHING_PRIORITY = 1.0
_TITLEDTHING_CHFREQ = "monthly"


class TitledThingSitemapRule(SitemapRule):
    __regid__ = "titledthing"
    query = "Any X WHERE X is TitledThing"
    priority = _TITLEDTHING_PRIORITY
    chfreq = _TITLEDTHING_CHFREQ


class NamedThingSitemapRule(SitemapRule):
    __regid__ = "namedthing"
    query = "Any X WHERE X is NamedThing"


#
# Mixin to parse Sitemap XML
#


class SitemapXMLCheckMixin:
    def assertSitemapContains(
        self,
        sitemap_data,
        expected_url,
        expected_lastmod=None,
        expected_priority=None,
        expected_chfreq=None,
    ):
        """Return True if the entry described by the parameters is found in the Sitemap data."""
        root_node = etree.fromstring(sitemap_data)
        for url_node in root_node.findall(_sitemap_tag("url")):
            if expected_lastmod is not None:
                expected_lastmod = (
                    expected_lastmod.strftime("%Y-%m-%d")
                    if isinstance(expected_lastmod, (date, datetime))
                    else str(expected_lastmod)
                )
                if url_node.findtext(_sitemap_tag("lastmod")) != expected_lastmod:
                    continue
            if (
                expected_priority is not None
                and url_node.findtext(_sitemap_tag("priority")) is not None
                and (
                    float(url_node.findtext(_sitemap_tag("priority")))
                    != float(expected_priority)
                )
            ):
                continue
            if expected_chfreq is not None and url_node.findtext(
                _sitemap_tag("changefreq")
            ) != str(expected_chfreq):
                continue
            if url_node.findtext(_sitemap_tag("loc")) == expected_url:
                break
        else:
            raise AssertionError(
                "Not found in Sitemap: {} (lastmod={!r}, priority={!r}, "
                "chfreq={!r})".format(
                    expected_url, expected_lastmod, expected_priority, expected_chfreq
                )
            )

    def assertSitemapLengthEqual(self, sitemap_data, expected_length):
        """Return True if the Sitemap data has the correct number of entries."""
        root_node = etree.fromstring(sitemap_data)
        length = len(list(root_node.findall(_sitemap_tag("url"))))
        if length != expected_length:
            raise AssertionError(
                f"Sitemap length is {length} (expected: {expected_length})"
            )

    def assertIndexContains(self, index_data, expected_url):
        """Return True if the given URL is found in the Sitemap index data."""
        root_node = etree.fromstring(index_data)
        for url_node in root_node.findall(_sitemap_tag("sitemap")):
            if url_node.findtext(_sitemap_tag("loc")) == expected_url:
                break
        else:
            raise AssertionError(f"Not found in Sitemap index: {expected_url}")


#
# Actual test cases
#


class SitemapViewTC(SitemapXMLCheckMixin, WebCWTC):
    """Test about the ``sitemapxml`` view."""

    def test_sitemap_view_one_entity(self):
        """Check that an XML sitemap is correctly generated with one entity."""
        with self.temporary_appobjects(TitledThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                thing = cnx.create_entity("TitledThing", title="A thing")
                cnx.commit()
            with self.admin_access.web_request() as req:
                thing = req.entity_from_eid(thing.eid)
                sitemap_data = req.view("sitemapxml").encode("utf-8")
                self.assertSitemapLengthEqual(sitemap_data, 2)  # Counting the root URL
                self.assertSitemapContains(
                    sitemap_data,
                    thing.absolute_url(),
                    thing.modification_date,
                    _TITLEDTHING_PRIORITY,
                    _TITLEDTHING_CHFREQ,
                )

    def test_sitemap_view_multiple_entities(self):
        """Check that an XML sitemap is correctly generated with multiple entities."""
        with self.temporary_appobjects(NamedThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                thing1 = cnx.create_entity("NamedThing", name="A thing")
                thing2 = cnx.create_entity("NamedThing", name="A second thing")
                cnx.commit()
            with self.admin_access.web_request() as req:
                thing1 = req.entity_from_eid(thing1.eid)
                thing2 = req.entity_from_eid(thing2.eid)
                sitemap_data = req.view("sitemapxml").encode("utf-8")
                self.assertSitemapLengthEqual(sitemap_data, 3)  # Counting the root URL
                self.assertSitemapContains(
                    sitemap_data, thing1.absolute_url(), thing1.modification_date
                )
                self.assertSitemapContains(
                    sitemap_data, thing2.absolute_url(), thing2.modification_date
                )

    def test_sitemap_view_multiple_etypes(self):
        """Check that an XML sitemap is correctly generated with multiple entities of multiple
        etypes."""
        with self.temporary_appobjects(TitledThingSitemapRule, NamedThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                thing0 = cnx.create_entity("TitledThing", title="A thing")
                thing1 = cnx.create_entity("NamedThing", name="A second thing")
                thing2 = cnx.create_entity("NamedThing", name="A third thing")
                cnx.commit()
            with self.admin_access.web_request() as req:
                thing0 = req.entity_from_eid(thing0.eid)
                thing1 = req.entity_from_eid(thing1.eid)
                thing2 = req.entity_from_eid(thing2.eid)
                sitemap_data = req.view("sitemapxml").encode("utf-8")
                self.assertSitemapLengthEqual(sitemap_data, 4)  # Counting the root URL
                self.assertSitemapContains(
                    sitemap_data,
                    thing0.absolute_url(),
                    thing0.modification_date,
                    _TITLEDTHING_PRIORITY,
                    _TITLEDTHING_CHFREQ,
                )
                self.assertSitemapContains(
                    sitemap_data, thing1.absolute_url(), thing1.modification_date
                )
                self.assertSitemapContains(
                    sitemap_data, thing2.absolute_url(), thing2.modification_date
                )


class GzippedSitemapViewTC(SitemapXMLCheckMixin, WebCWTC):
    """Test about the ``sitemapxmlgz`` view."""

    def test_gz_sitemap_view(self):
        """Check that a Gzipped XML sitemap is correctly generated."""
        with self.temporary_appobjects(TitledThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                thing = cnx.create_entity("TitledThing", title="A thing")
                cnx.commit()
            with self.admin_access.web_request() as req:
                thing = req.entity_from_eid(thing.eid)
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    f.write(req.view("sitemapxmlgz"))
                sitemap_data = BytesIO()
                with gzip.open(f.name, mode="rb") as gz:
                    sitemap_data.write(gz.read())
                os.remove(f.name)
                sitemap_data = sitemap_data.getvalue()
                self.assertSitemapLengthEqual(sitemap_data, 2)  # Counting the root URL
                self.assertSitemapContains(
                    sitemap_data,
                    thing.absolute_url(),
                    thing.modification_date,
                    _TITLEDTHING_PRIORITY,
                    _TITLEDTHING_CHFREQ,
                )


class ZippedSitemapsViewTC(SitemapXMLCheckMixin, WebCWTC):
    """Test about the ``massive-sitemaps`` view."""

    def create_many_things(self, cnx, n):
        """Create as many as ``n`` ``NamedThing`` using the given connection."""
        for i in range(n):
            cnx.create_entity("NamedThing", name=f"Thing {i}")
            if i % 1000 == 0:
                print(i)
        cnx.commit()

    def extract_file(self, fname):
        """Extract the given ZIP file and return list of extracted filenames."""
        tmpdir = tempfile.gettempdir()
        sitemaps = []
        with zipfile.ZipFile(fname, mode="r") as zf:
            zf.extractall(tmpdir)
            for gz_fname in zf.namelist():
                sitemap_data = BytesIO()
                with gzip.open(os.path.join(tmpdir, gz_fname), mode="rb") as gz:
                    sitemap_data.write(gz.read())
                sitemaps.append(sitemap_data)
        return sitemaps

    def clean(self, zip_fname, gz_number):
        """Remove temporary files."""
        os.remove(zip_fname)
        os.remove(os.path.join(tempfile.gettempdir(), "sitemap.xml.gz"))
        for i in range(1, gz_number):
            os.remove(os.path.join(tempfile.gettempdir(), f"sitemap{i}.xml.gz"))

    def test_zip_sitemaps_view_one_file(self):
        """Check that a bunch of XML sitemaps is correctly generated and archived in a ZIP file."""
        with self.temporary_appobjects(TitledThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                thing = cnx.create_entity("TitledThing", title="A thing")
                cnx.commit()
            with self.admin_access.web_request() as req:
                thing = req.entity_from_eid(thing.eid)
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    f.write(req.view("massive-sitemaps"))
                sitemaps = self.extract_file(f.name)
                os.remove(f.name)
                self.assertEqual(len(sitemaps), 2)
                index_data = sitemaps[0].getvalue()
                self.assertIndexContains(index_data, req.build_url("sitemap1.xml.gz"))
                sitemap_data = sitemaps[1].getvalue()
                self.assertSitemapLengthEqual(sitemap_data, 2)  # Counting the root URL
                self.assertSitemapContains(
                    sitemap_data,
                    thing.absolute_url(),
                    thing.modification_date,
                    _TITLEDTHING_PRIORITY,
                    _TITLEDTHING_CHFREQ,
                )

    def test_zip_sitemaps_view_one_full_file(self):
        """Check that an XML sitemap is correctly generated, gzipped, and archived in a ZIP file
        with a Sitemap index file."""
        with self.temporary_appobjects(NamedThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                self.create_many_things(cnx, 49999)
            with self.admin_access.web_request() as req:
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    f.write(req.view("massive-sitemaps"))
                sitemaps = self.extract_file(f.name)
                self.clean(f.name, len(sitemaps))
                self.assertEqual(len(sitemaps), 2)
                index_data = sitemaps[0].getvalue()
                self.assertIndexContains(index_data, req.build_url("sitemap1.xml.gz"))
                sitemap_data = sitemaps[1].getvalue()
                self.assertSitemapLengthEqual(
                    sitemap_data, 50000
                )  # Counting the root URL

    def test_zip_sitemaps_view_two_files(self):
        """Check that a bunch of XML sitemaps are correctly generated, gzipped, and archived in a
        ZIP file with a Sitemap index file."""
        with self.temporary_appobjects(NamedThingSitemapRule):
            with self.admin_access.repo_cnx() as cnx:
                self.create_many_things(cnx, 50000)
            with self.admin_access.web_request() as req:
                with tempfile.NamedTemporaryFile(delete=False) as f:
                    f.write(req.view("massive-sitemaps"))
                sitemaps = self.extract_file(f.name)
                self.clean(f.name, len(sitemaps))
                self.assertEqual(len(sitemaps), 3)
                index_data = sitemaps[0].getvalue()
                self.assertIndexContains(index_data, req.build_url("sitemap1.xml.gz"))
                self.assertIndexContains(index_data, req.build_url("sitemap2.xml.gz"))
                sitemap_data = sitemaps[1].getvalue()
                self.assertSitemapLengthEqual(
                    sitemap_data, 50000
                )  # Counting the root URL
                sitemap_data = sitemaps[2].getvalue()
                self.assertSitemapLengthEqual(sitemap_data, 1)



# copyright 2012-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-seo views/forms/actions/components for web ui"""

from io import StringIO
from datetime import date
import gzip
import os
import tempfile
import urllib.parse
import zipfile

from cubicweb import Unauthorized
from cubicweb_web.view import View
from cubicweb.predicates import none_rset
from cubicweb.appobject import AppObject
from cubicweb_web import httpcache
from cubicweb_web.views.urlrewrite import SimpleReqRewriter, rgx


class SeoReqRewriter(SimpleReqRewriter):
    rules = [
        (rgx("/robots.txt$"), dict(vid="robotstxt")),
        (rgx("/sitemap.xml$"), dict(vid="sitemapxml")),
    ]


class RobotsRule(AppObject):
    """abstract base class for robots rules

    A custom rule should be built by extending this class. For instance::

      from cubicweb_seo.views import RobotsRule

      class VcsfileRobotsRule(RobotsRule):
          __regid__ = 'vcsfile'
          STANZA = ['User-Agent: *',
                    'Disallow: /versioncontent',
                    'Disallow: /testexecution',
                    'Disallow: /versionedfile',
                    'Disallow: /deletedversioncontent']

    To get more control, override the items() method.
    """

    __registry__ = "robotstxt"
    __abstract__ = True
    STANZA = []

    def stanza(self):
        return "\n".join(self.STANZA)


class CwRobotsRule(RobotsRule):
    __regid__ = "cubicweb.admin"
    STANZA = [
        "User-Agent: *",
        "Disallow: /schema",
        "Disallow: /cwetype",
        "Disallow: /cwrtype",
        "Disallow: /search",
        "Disallow: /login",
        "Disallow: /cwuser",
        "Disallow: /cwgroup",
        "Disallow: /cwsource",
        "Disallow: /_registry",
        "Disallow: /siteconfig",
        "Disallow: /siteinfo",
        "Disallow: /notfound",
        "Disallow: /processinfo",
        "Disallow: /changelog",
        "Disallow: /add/",
    ]


class RobotsTxt(View):
    """A view that implements the http://robotstxt.org standard"""

    __regid__ = "robotstxt"
    content_type = "text/plain"
    templatable = False
    __select__ = none_rset()

    def call(self):
        self.w(f"Sitemap: {self._cw.build_url('sitemap.xml')}\n\n")
        if (
            "robotstxt" in self._cw.vreg
        ):  # XXX or make sure it exists by creating it in this cube...
            for rule in self._cw.vreg["robotstxt"].possible_objects(self._cw):
                self.w(rule.stanza())
                self.w("\n")


CHANGEFREQS = frozenset(
    ("always", "hourly", "daily", "weekly", "monthly", "yearly", "never")
)


class SitemapRule(AppObject):
    """abstract base class for sitemap rules

    A custom rule should be built by extending this class. For instance::

      from cubicweb_seo.views import SitemapRule

      class CardSitemapRule(SitemapRule):
          __regid__ = 'card'
          query = 'Any X WHERE X is Card'
          priority = 1.0
          chfreq = 'monthly'

    To get more control, override the items() method.
    """

    __registry__ = "sitemap"
    __abstract__ = True
    query = None
    priority = None  # optional
    chfreq = None  # optional

    def items(self):
        rset = self._cw.execute(self.query)
        for item in rset.entities():
            yield (
                item.absolute_url(),
                item.modification_date,
                self.chfreq,
                self.priority,
            )


class Sitemaps(View):
    """A view that implements the http://sitemaps.org standard

    Content types must append rules to the sitemap registry in order to appear
    in the list at /sitemap.xml.
    """

    __regid__ = "sitemapxml"
    content_type = "text/xml"
    templatable = False
    __select__ = none_rset()

    def add_url(self, url, lastmod=None, chfreq=None, priority=None):
        self.w(f"<url><loc>{url}</loc>")
        if lastmod:
            self.w(f"<lastmod>{lastmod.strftime('%Y-%m-%d')}</lastmod>")
        if chfreq:
            if chfreq in CHANGEFREQS:
                self.w(f"<changefreq>{chfreq}</changefreq>")
            else:
                self.error(
                    "url %s gave %s that is not a valid changefreq according to http://sitemaps.org"
                    % (url, chfreq)
                )
        if priority:
            if 0 <= float(priority) <= 1:
                self.w(f"<priority>{priority:.2f}</priority>")
            else:
                self.error(
                    "url %s gave %s which is not a valid priority according to http://sitemaps.org"
                    % (url, priority)
                )
        self.w("</url>")

    def call(self):
        # XXX have a look at databnf/file/stable/sitemap.py for cases where the sitemap file is
        # too big (>10Mb) or has too many entries (>50k)
        self.w(f'<?xml version="1.0" encoding="{self._cw.encoding}"?>\n')
        self.w('<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">')
        self.add_url(self._cw.base_url())
        if (
            "sitemap" in self._cw.vreg
        ):  # XXX or make sure it exists by creating it in this cube...
            count = 0
            for rule in self._cw.vreg["sitemap"].possible_objects(self._cw):
                try:
                    for item in rule.items():
                        self.add_url(*item)
                        count += 1
                        if count > 50000:
                            self.error(
                                "too many items for this sitemap (limit is 50,000)"
                            )
                            break
                    if count > 50000:
                        break
                except Unauthorized as exc:
                    self.info(
                        f"sitemap rule {rule.__regid__} raised Unauthorized: {exc}"
                    )
                except Exception:
                    self.exception(f"sitemap rule {rule.__regid__} raised exception")
        else:
            self.debug("no sitemap registry found")
        self.w("</urlset>")


class GzippedSitemap(Sitemaps):
    """Allows to download a Gzipped sitemap.xml."""

    __regid__ = "sitemapxmlgz"
    content_type = "application/gzip"
    binary = True
    http_cache_manager = httpcache.EntityHTTPCacheManager
    add_to_breadcrumbs = False

    def add_url(self, buf, url, lastmod=None, chfreq=None, priority=None):
        buf.write("<url>")
        buf.write(f"<loc>{urllib.parse.quote(url.encode('utf-8'), safe=':/')}</loc>")
        if lastmod:
            buf.write(f"<lastmod>{lastmod.strftime('%Y-%m-%d')}</lastmod>")
        if chfreq:
            if chfreq in CHANGEFREQS:
                buf.write(f"<changefreq>{chfreq}</changefreq>")
            else:
                self.error(
                    "url %s gave %s which is not a valid changefreq according to "
                    "http://sitemaps.org" % (url, chfreq)
                )
        if priority:
            if 0 <= float(priority) <= 1:
                buf.write(f"<priority>{priority}</priority>")
            else:
                self.error(
                    "url %s gave %s which is not a valid priority according to "
                    "http://sitemaps.org" % (url, priority)
                )
        buf.write("</url>")

    def call(self):
        buf = StringIO()
        buf.write(f'<?xml version="1.0" encoding="{self._cw.encoding}"?>\n')
        buf.write('<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">')
        self.add_url(buf, self._cw.base_url())
        if (
            "sitemap" in self._cw.vreg
        ):  # XXX or make sure it exists by creating it in this cube...
            count = 0
            for rule in self._cw.vreg["sitemap"].possible_objects(self._cw):
                try:
                    for item in rule.items():
                        self.add_url(buf, *item)
                        count += 1
                        if count > 50000:
                            self.error(
                                "too many items for this sitemap (limit is 50,000)"
                            )
                            break
                    if count > 50000:
                        break
                except Unauthorized as exc:
                    self.info(
                        f"sitemap rule {rule.__regid__} raised Unauthorized: {exc}"
                    )
                except Exception:
                    self.exception(f"sitemap rule {rule.__regid__} raised exception")
        else:
            self.debug("no sitemap registry found")
        buf.write("</urlset>")
        with tempfile.NamedTemporaryFile(delete=False) as f:
            with gzip.open(f.name, mode="wb") as gz:
                gz.write(buf.getvalue().encode("utf-8"))
        self._cw.set_content_type(
            self.content_type, filename="sitemap.xml.gz", disposition="attachment"
        )
        with open(f.name, "rb") as gz:
            self.w(gz.read())
        os.remove(f.name)


class ZippedMultipleSitemaps(GzippedSitemap):
    """Allows to download a ZIP archive containing multiple Sitemap files and a Sitemap index.

    This view should be used when number of URLs in sitemap will exceed the 50,000 limit.
    Thus URLs must be dispatched among multiple Sitemap files and a Sitemap index file must be
    created.

    Sitemap files will always be gzipped in this case.
    """

    __regid__ = "massive-sitemaps"
    content_type = "application/zip"
    binary = True
    http_cache_manager = httpcache.EntityHTTPCacheManager
    add_to_breadcrumbs = False

    def new_sitemap_buffer(self, index=False):
        """Create a new buffer storing content about a Sitemap file in memory.

        If ``index`` is True, start a buffer for a Sitemap index file."""
        buf = StringIO()
        buf.write(f'<?xml version="1.0" encoding="{self._cw.encoding}"?>\n')
        tagname = "urlset" if not index else "sitemapindex"
        buf.write(f'<{tagname} xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">')
        return buf

    def save_partial_sitemap(self, sitemap_buf, fname, index_buf):
        """Save content in ``sitemap_buf`` into ``fname``, and add an index entry in
        ``index_buf``."""
        sitemap_buf.write("</urlset>")
        with gzip.open(fname, mode="wb") as gz:
            gz.write(sitemap_buf.getvalue().encode("utf-8"))
        index_buf.write("<sitemap>")
        index_buf.write(f"<loc>{self._cw.build_url(os.path.basename(fname))}</loc>")
        index_buf.write(f"<lastmod>{date.today().strftime('%y-%m-%d')}</lastmod>")
        index_buf.write("</sitemap>")

    def save_sitemap_index(self, index_buf, index_fname):
        """Save content in ``sitemap_buf`` into ``fname``, and add an index entry in
        ``index_buf``."""
        index_buf.write("</sitemapindex>")
        with gzip.open(index_fname, mode="wb") as gz:
            gz.write(index_buf.getvalue().encode("utf-8"))

    def call(self):
        index_fname = os.path.join(tempfile.gettempdir(), "sitemap.xml.gz")
        fname_templ = os.path.join(tempfile.gettempdir(), "sitemap{0}.xml.gz")
        buf = self.new_sitemap_buffer()
        index_buf = self.new_sitemap_buffer(index=True)
        self.add_url(buf, self._cw.base_url())
        if (
            "sitemap" in self._cw.vreg
        ):  # XXX or make sure it exists by creating it in this cube...
            entries = file_number = 1
            for rule in self._cw.vreg["sitemap"].possible_objects(self._cw):
                try:
                    for item in rule.items():
                        entries += 1
                        if entries > 50000:
                            self.save_partial_sitemap(
                                buf, fname_templ.format(file_number), index_buf
                            )
                            file_number += 1
                            buf = self.new_sitemap_buffer()
                            entries = 1
                        self.add_url(buf, *item)
                except Unauthorized as exc:
                    self.info(
                        f"sitemap rule {rule.__regid__} raised Unauthorized: {exc}"
                    )
                except Exception:
                    self.exception(f"sitemap rule {rule.__regid__} raised exception")
            self.save_partial_sitemap(buf, fname_templ.format(file_number), index_buf)
            self.save_sitemap_index(index_buf, index_fname)
        else:
            self.debug("no sitemap registry found")
        with tempfile.NamedTemporaryFile(delete=False) as f:
            with zipfile.ZipFile(f.name, mode="w") as zf:
                zf.write(index_fname, os.path.basename(index_fname))
                for i in range(1, file_number + 1):
                    fname = fname_templ.format(i)
                    zf.write(fname, os.path.basename(fname))
        self._cw.set_content_type(
            self.content_type, filename="sitemaps.zip", disposition="attachment"
        )
        with open(f.name, "rb") as zf:
            self.w(zf.read())
        os.remove(f.name)
        for i in range(1, file_number + 1):
            os.remove(fname_templ.format(i))



from yams.buildobjs import EntityType, String


class NamedThing(EntityType):
    name = String(maxsize=42, required=True)


class TitledThing(EntityType):
    title = String(maxsize=42, required=True)
